"""Utility functions for transformers.

This module provides common utility functions used across different transformers
for text processing and URI/name building.
"""

import re


def process_text(text: str, max_length: int = 100000) -> str:
    """Process and sanitize text for storage.

    This function processes text by:
    1. Truncating it to a maximum length
    2. Removing HTML tags
    3. Converting to a JSON-safe string

    Args:
        text (str): The text to process.
        max_length (int, optional): Maximum length of text to keep. Defaults to 100000.

    Returns:
        str: The processed text, truncated and sanitized.
    """
    if len(text) > max_length:
        text = text[:max_length]

    # Remove HTML tags
    text = re.sub(r"\s+", " ", re.sub(r"<[^>]+>", "", text)).strip()

    return text


def build_phoenix_uri(connector_name: str, connector_type: str, *args: str) -> str:
    """Build the URI for a Phoenix entity.

    This function constructs a URI path by joining the connector name, type,
    and additional path components.

    Args:
        connector_name (str): Name of the connector.
        connector_type (str): Type of the connector.
        *args (str): Additional path components to append.

    Returns:
        str: The constructed URI path, starting with a forward slash.

    Example:
        >>> build_phoenix_uri("myconnector", "sql", "db", "schema")
        '/myconnector/sql/db/schema'
    """
    return f"/{connector_name}/{connector_type}/{'/'.join(args)}"


def build_atlas_qualified_name(connection_qualified_name: str, *args: str) -> str:
    """Build the qualified name for an Atlas entity.

    This function constructs a qualified name by joining the connection qualified name
    and additional name components.

    Args:
        connection_qualified_name (str): Base qualified name for the connection.
        *args (str): Additional name components to append.

    Returns:
        str: The constructed qualified name with components joined by forward slashes.

    Example:
        >>> build_atlas_qualified_name("tenant/connector/1", "db", "schema")
        'tenant/connector/1/db/schema'
    """
    return f"{connection_qualified_name}/{'/'.join(args)}"
